@@ -28,51 +28,34 @@
 
 from __future__ import print_function
 
-import argparse
-import cmd
 import logging
 import os
 import re
 import shlex
 import sys
+import time
 
-from kazoo.client import KazooClient
-from kazoo.exceptions import NotEmptyError
-from kazoo.security import make_acl, make_digest_acl
+from kazoo.exceptions import NoNodeError, NotEmptyError
 
+from .acl import ACLReader
+from .augumented_client import AugumentedClient
+from .augumented_cmd import AugumentedCmd
 from .copy import copy, CopyError
 from .watch_manager import get_watch_manager
+from .util import pretty_bytes, to_bool
 
 
-class ShellParser(argparse.ArgumentParser):
-    def error(self, message):
-        raise Exception(message)
-
-
-class Shell(cmd.Cmd):
-    curdir = '/'
-
+class Shell(AugumentedCmd):
     def __init__(self, hosts=[], timeout=10):
-        cmd.Cmd.__init__(self)
+        AugumentedCmd.__init__(self, ".kz-shell-history")
         self._hosts = hosts
         self._connect_timeout = timeout
-        self._setup_readline()
-
         self._zk = None
         self._read_only = False
         self.connected = False
 
-        if len(self._hosts) > 0:
-            self._connect(self._hosts)
-
-        if not self.connected:
-            self._update_curdir('/')
-
-    def emptyline(self):
-        pass
-
-    def run(self):
-        self.cmdloop("")
+        if len(self._hosts) > 0: self._connect(self._hosts)
+        if not self.connected: self.update_curdir("/")
 
     def connected(f):
         def wrapped(self, args):
@@ -81,56 +64,27 @@ class Shell(cmd.Cmd):
           print("Not connected.")
         return wrapped
 
-    def interruptible(f):
-        def wrapped(self, args):
-            try:
-                f(self, args)
-            except KeyboardInterrupt:
-                pass
-        return wrapped
-
-    def ensure_params(expected_params):
-        def wrapper(f):
-            parser = ShellParser()
-            for p, optional in expected_params:
-                if optional is True:
-                    parser.add_argument(p)
-                elif optional is False:
-                    parser.add_argument(p, nargs="?", default="")
-                elif optional is "+":
-                    parser.add_argument(p, nargs="+")
-
-            def wrapped(self, args):
-                try:
-                    params = parser.parse_args(shlex.split(args))
-                    return f(self, params)
-                except Exception as ex:
-                    valid_params = " ".join(
-                        e[0] if e[1] else "<%s>" % (e[0]) for e in expected_params)
-                    print("Wrong params: %s. Expected: %s" % (str(ex), valid_params))
-            return wrapped
-        return wrapper
-
     def check_path_exists(f):
         def wrapped(self, params):
             path = params.path
-            params.path = self._abspath(path if path not in ["", "."] else self.curdir)
+            params.path = self.abspath(path if path not in ["", "."] else self.curdir)
             if self._zk.exists(params.path):
                 return f(self, params)
             print("Path %s doesn't exist" % (path))
+            return False
         return wrapped
 
     def check_path_absent(f):
         def wrapped(self, params):
             path = params.path
-            params.path = self._abspath(path if path != '' else self.curdir)
+            params.path = self.abspath(path if path != '' else self.curdir)
             if not self._zk.exists(params.path):
                 return f(self, params)
             print("Path %s already exists" % (path))
         return wrapped
 
     @connected
-    @ensure_params([("scheme", True), ("credential", True)])
+    @AugumentedCmd.ensure_params([("scheme", True), ("credential", True)])
     def do_add_auth(self, params):
         self._zk.add_auth(params.scheme, params.credential)
 
@@ -142,13 +96,11 @@ example:
   add_auth digest super:s3cr3t
 """)
 
-    class BadACL(Exception): pass
-
     @connected
-    @ensure_params([("path", True), ("acls", "+")])
+    @AugumentedCmd.ensure_params([("path", True), ("acls", "+")])
     @check_path_exists
     def do_set_acls(self, params):
-        acls = self._extract_acls(params.acls)
+        acls = ACLReader.extract(params.acls)
         try:
             self._zk.set_acls(params.path, acls)
         except Exception as ex:
@@ -163,45 +115,8 @@ example:
   set_acls /some/path world:anyone:r username_password:user:p@ass0rd:cdrwa
 """)
 
-    def _extract_acls(self, acls):
-        return map(self._extract_acl, acls)
-
-    valid_schemes = [
-        "world",
-        "auth",
-        "digest",
-        "host",
-        "ip",
-        "username_password",  # internal-only: gen digest from user:password
-    ]
-
-    def _extract_acl(self, acl):
-        try:
-            scheme, rest = acl.split(":", 1)
-            credential = ":".join(rest.split(":")[0:-1])
-            cdrwa = rest.split(":")[-1]
-        except ValueError:
-            raise self.BadACL("Bad ACL: %s. Format is scheme:id:perms" % (acl))
-
-        if scheme not in self.valid_schemes:
-            raise self.BadACL("Invalid scheme: %s" % (acl))
-
-        create = True if "c" in cdrwa else False
-        read = True if "r" in cdrwa else False
-        write = True if "w" in cdrwa else False
-        delete = True if "d" in cdrwa else False
-        admin = True if "a" in cdrwa else False
-
-        if scheme == "username_password":
-            username, password = credential.split(":", 1)
-            return make_digest_acl(username, password, read, write, create,
-                                   delete, admin)
-        else:
-            return make_acl(scheme, credential, read, write,
-                            create, delete, admin)
-
     @connected
-    @ensure_params([("path", True)])
+    @AugumentedCmd.ensure_params([("path", True)])
     @check_path_exists
     def do_get_acls(self, params):
         print(self._zk.get_acls(params.path)[0])
@@ -216,22 +131,22 @@ example:
 """)
 
     @connected
-    @ensure_params([("path", False), ("watch", False)])
+    @AugumentedCmd.ensure_params([("path", False), ("watch", False)])
     @check_path_exists
     def do_ls(self, params):
-        if params.watch.lower() == "true":
-            znodes = self._zk.get_children(params.path, watch=self._default_watcher)
-        else:
-            znodes = self._zk.get_children(params.path)
+        kwargs = {"watch": self.watcher} if to_bool(params.watch) else {}
+        znodes = self._zk.get_children(params.path, **kwargs)
         print(" ".join(znodes))
 
     def complete_ls(self, cmd_param_text, full_cmd, start_idx, end_idx):
         return self._complete_path(cmd_param_text, full_cmd)
 
     @connected
-    @ensure_params([("command", True),
+    @AugumentedCmd.interruptible
+    @AugumentedCmd.ensure_params([("command", True),
                     ("path", True),
-                    ("debug", False)])
+                    ("debug", False),
+                    ("sleep", False)])
     @check_path_exists
     def do_watch(self, params):
         wm = get_watch_manager(self._zk)
@@ -240,7 +155,20 @@ example:
         elif params.command == "stop":
             wm.remove(params.path)
         elif params.command == "stats":
-            wm.stats(params.path)
+            repeat = 1
+            sleep = 1
+            try:
+                repeat = int(params.debug)
+                sleep = int(params.sleep)
+            except ValueError: pass
+            if repeat == 0:
+                while True:
+                    wm.stats(params.path)
+                    time.sleep(sleep)
+            else:
+                for i in xrange(0, repeat):
+                    wm.stats(params.path)
+                    time.sleep(sleep)
         else:
             print("watch <start|stop> <path> [verbose]")
 
@@ -249,12 +177,12 @@ example:
 Recursively watch for all changes under a path.
 
 examples:
-  watch start /foo/bar
+  watch start /foo/bar [debug]
   watch stop /foo/bar
-  watch stats /foo/bar
+  watch stats /foo/bar [repeatN] [sleepN]
 """)
 
-    @ensure_params([("src", True), ("dst", True),
+    @AugumentedCmd.ensure_params([("src", True), ("dst", True),
                     ("recursive", False), ("overwrite", False),
                     ("verbose", False)])
     def do_cp(self, params):
@@ -275,8 +203,8 @@ example:
 """)
 
     @connected
-    @interruptible
-    @ensure_params([("path", False), ("max_depth", False)])
+    @AugumentedCmd.interruptible
+    @AugumentedCmd.ensure_params([("path", False), ("max_depth", False)])
     @check_path_exists
     def do_tree(self, params):
         max_depth = 0
@@ -285,7 +213,9 @@ example:
         except ValueError: pass
 
         print(".")
-        self._print_tree(params.path, 0, max_depth)
+        self._zk.tree(params.path,
+                      max_depth,
+                      lambda c,l: print(u"%s├── %s" % (u"│   " * l, c)))
 
     def complete_tree(self, cmd_param_text, full_cmd, start_idx, end_idx):
         return self._complete_path(cmd_param_text, full_cmd)
@@ -308,17 +238,21 @@ example:
    ├── bar
 """)
 
-    def _print_tree(self, path, level, max_depth):
-        for c in self._zk.get_children(path):
-            print(u"%s├── %s" % (u"│   " * level, c))
-            if max_depth == 0 or level + 1 < max_depth:
-                self._print_tree(u"%s/%s" % (path, c), level + 1, max_depth)
+    @connected
+    @AugumentedCmd.ensure_params([("path", False)])
+    @check_path_exists
+    def do_du(self, params):
+        print(pretty_bytes(self._zk.du(params.path)))
 
     @connected
-    @ensure_params([("path", True), ("match", True)])
+    @AugumentedCmd.ensure_params([("path", True), ("match", True)])
     @check_path_exists
     def do_find(self, params):
-        self._full_match(params.path, params.match, True, 0)
+        self._zk.find(params.path,
+                      params.match,
+                      True,
+                      0,
+                      lambda p: print(p))
 
     def complete_find(self, cmd_param_text, full_cmd, start_idx, end_idx):
         return self._complete_path(cmd_param_text, full_cmd)
@@ -336,10 +270,14 @@ example:
 """)
 
     @connected
-    @ensure_params([("path", True), ("match", True)])
+    @AugumentedCmd.ensure_params([("path", True), ("match", True)])
     @check_path_exists
     def do_ifind(self, params):
-        self._full_match(params.path, params.match, True, re.IGNORECASE)
+        self._zk.find(params.path,
+                      params.match,
+                      True,
+                      re.IGNORECASE,
+                      lambda p: print(p))
 
     def complete_ifind(self, cmd_param_text, full_cmd, start_idx, end_idx):
         return self._complete_path(cmd_param_text, full_cmd)
@@ -356,25 +294,16 @@ example:
   /copy/Foo
 """)
 
-    def _full_match(self, path, match, check_match, flags):
-        for c in self._zk.get_children(path):
-            check = check_match
-            full_path = os.path.join(path, c)
-            if not check:
-                print(full_path)
-            else:
-                check = not re.search(match, full_path, flags)
-                if not check:
-                    print(full_path)
-
-            self._full_match(full_path, match, check, flags)
-
     @connected
-    @ensure_params([("path", True), ("content", True), ("show_matches", False)])
+    @AugumentedCmd.ensure_params([("path", True), ("content", True), ("show_matches", False)])
     @check_path_exists
     def do_grep(self, params):
         show_matches = params.show_matches.lower() == "true"
-        self._full_grep(params.path, params.content, show_matches, 0)
+        self._.zk.grep(params.path,
+                       params.content,
+                       show_matches,
+                       0,
+                       lambda p: print(p))
 
     def complete_grep(self, cmd_param_text, full_cmd, start_idx, end_idx):
         return self._complete_path(cmd_param_text, full_cmd)
@@ -390,11 +319,15 @@ example:
 """)
 
     @connected
-    @ensure_params([("path", True), ("content", True), ("show_matches", False)])
+    @AugumentedCmd.ensure_params([("path", True), ("content", True), ("show_matches", False)])
     @check_path_exists
     def do_igrep(self, params):
         show_matches = params.show_matches.lower() == "true"
-        self._full_grep(params.path, params.content, show_matches, re.IGNORECASE)
+        self._zk.grep(params.path,
+                      params.content,
+                      show_matches,
+                      re.IGNORECASE,
+                      lambda p: print(p))
 
     def complete_igrep(self, cmd_param_text, full_cmd, start_idx, end_idx):
         return self._complete_path(cmd_param_text, full_cmd)
@@ -409,38 +342,21 @@ example:
   /copy/passwd: unbound:x:992:991:Unbound DNS resolver:/etc/unbound:/sbin/nologin
 """)
 
-    def _full_grep(self, path, content, show_matches, flags):
-        for c in self._zk.get_children(path):
-            full_path = os.path.join(path, c)
-            value, _ = self._zk.get(full_path)
-
-            if show_matches:
-                for line in value.split("\n"):
-                    if re.search(content, line, flags):
-                        print("%s: %s" % (full_path, line))
-            else:
-                if re.search(content, value, flags):
-                    print(full_path)
-
-            self._full_grep(full_path, content, show_matches, flags)
-
     @connected
-    @ensure_params([("path", True)])
+    @AugumentedCmd.ensure_params([("path", True)])
     @check_path_exists
     def do_cd(self, params):
-        self._update_curdir(params.path)
+        self.update_curdir(params.path)
 
     def complete_cd(self, cmd_param_text, full_cmd, start_idx, end_idx):
         return self._complete_path(cmd_param_text, full_cmd)
 
     @connected
-    @ensure_params([("path", True), ("watch", False)])
+    @AugumentedCmd.ensure_params([("path", True), ("watch", False)])
     @check_path_exists
     def do_get(self, params):
-        if params.watch.lower() == "true":
-            value, stat = self._zk.get(params.path, watch=self._default_watcher)
-        else:
-            value, stat = self._zk.get(params.path)
+        kwargs = {"watch": self.watcher} if to_bool(params.watch) else {}
+        value, stat = self._zk.get(params.path, **kwargs)
         print(value)
 
     def complete_get(self, cmd_param_text, full_cmd, start_idx, end_idx):
@@ -463,13 +379,11 @@ example:
 """)
 
     @connected
-    @ensure_params([("path", True), ("watch", False)])
+    @AugumentedCmd.ensure_params([("path", True), ("watch", False)])
     @check_path_exists
     def do_exists(self, params):
-        if params.watch.lower() == "true":
-            stat = self._zk.exists(params.path, watch=self._default_watcher)
-        else:
-            stat = self._zk.exists(params.path)
+        kwargs = {"watch": self.watcher} if to_bool(params.watch) else {}
+        stat = self._zk.exists(params.path, watch=self.watcher, **kwargs)
         print(stat)
 
     def complete_exists(self, cmd_param_text, full_cmd, start_idx, end_idx):
@@ -491,11 +405,11 @@ example:
   WatchedEvent(type='DELETED', state='CONNECTED', path=u'/foo')
 """)
 
-    def _default_watcher(self, watched_event):
+    def watcher(self, watched_event):
         print((str(watched_event)))
 
     @connected
-    @ensure_params([("path", True),
+    @AugumentedCmd.ensure_params([("path", True),
                     ("value", True),
                     ("ephemeral", False),
                     ("sequence", False),
@@ -544,7 +458,7 @@ example:
 """)
 
     @connected
-    @ensure_params([("path", True), ("value", True)])
+    @AugumentedCmd.ensure_params([("path", True), ("value", True)])
     @check_path_exists
     def do_set(self, params):
         self._zk.set(params.path, params.value)
@@ -561,7 +475,7 @@ example:
 """)
 
     @connected
-    @ensure_params([("path", True)])
+    @AugumentedCmd.ensure_params([("path", True)])
     @check_path_exists
     def do_rm(self, params):
         try:
@@ -573,17 +487,10 @@ example:
         return self._complete_path(cmd_param_text, full_cmd)
 
     @connected
-    @ensure_params([("path", True)])
+    @AugumentedCmd.ensure_params([("path", True)])
     @check_path_exists
     def do_rmr(self, params):
-        self._delete_recursive(params.path)
-
-    def _delete_recursive(self, path):
-        for c in self._zk.get_children(path):
-            child_path = os.path.join(path, c)
-            self._delete_recursive(child_path)
-
-        self._zk.delete(path)
+        self._zk.rmr(params.path)
 
     def complete_rmr(self, cmd_param_text, full_cmd, start_idx, end_idx):
         return self._complete_path(cmd_param_text, full_cmd)
@@ -597,12 +504,12 @@ example:
 """)
 
     @connected
-    @ensure_params([("path", True)])
+    @AugumentedCmd.ensure_params([("path", True)])
     @check_path_exists
     def do_sync(self, params):
         self._zk.sync(params.path)
 
-    @ensure_params([("hosts", True)])
+    @AugumentedCmd.ensure_params([("hosts", True)])
     def do_connect(self, params):
         self._connect(params.hosts.split(","))
 
@@ -617,7 +524,7 @@ example:
     @connected
     def do_disconnect(self, args):
         self._disconnect()
-        self._update_curdir('/')
+        self.update_curdir("/")
 
     def help_disconnect(self):
         print("""
@@ -641,75 +548,25 @@ example:
         self._exit(False)
 
     def _disconnect(self):
-        if self._zk:
-            try:
-                self._zk.stop()
-            except:
-                pass
+        if self._zk: return
+
+        try: self._zk.stop()
+        except Exception: pass
 
     def _connect(self, hosts):
         self._disconnect()
-        self._zk = KazooClient(hosts=",".join(hosts), read_only=self._read_only)
+        self._zk = AugumentedClient(",".join(hosts),
+                                    read_only=self._read_only)
         try:
             self._zk.start(timeout=self._connect_timeout)
             self.connected = True
-        except Exception as ex:
-            print("Failed to connect: %s" % (ex))
-        self._update_curdir('/')
+        except Exception as ex: print("Failed to connect: %s" % (ex))
 
-    def _update_curdir(self, dirpath):
-        if dirpath == '..':
-            if self.curdir == '/':
-                dirpath = '/'
-            else:
-                dirpath = os.path.dirname(self.curdir)
-        elif not dirpath.startswith('/'):
-            prefix = self.curdir
-            if prefix != '/':
-                prefix += '/'
-            dirpath = prefix + dirpath
+        self.update_curdir("/")
 
-        if dirpath != '/' and not self._zk.exists(dirpath):
-            print("Path %s doesn't exist." % dirpath)
-        else:
-            self.curdir = dirpath
-            self.prompt = "(%s) %s> " % (self._state(), dirpath)
-
-    def _state(self):
-        if self._zk:
-            return self._zk.state
-        return "DISCONNECTED"
-
-    def _exit(self, newline=True):
-        if newline:
-            print("")
-        sys.exit(0)
-
-    def _abspath(self, path):
-        if path != '/': path = path.rstrip('/')
-
-        if path == '..':
-            return os.path.dirname(self.curdir)
-        elif path.startswith('/'):
-            return path
-        elif self.curdir == '/':
-            return "/%s" % (path)
-        else:
-            return "%s/%s" % (self.curdir, path)
-
-    def _setup_readline(self):
-        try:
-            import readline
-            import atexit
-        except ImportError:
-            return
-
-        histfile = os.path.join(os.environ['HOME'], '.kz-shell-history')
-        try:
-            readline.read_history_file(histfile)
-        except IOError:
-            pass
-        atexit.register(readline.write_history_file, histfile)
+    @property
+    def state(self):
+        return "(%s) " % (self._zk.state if self._zk else "DISCONNECTED")
 
     def _complete_path(self, cmd_param_text, full_cmd):
         pieces = shlex.split(full_cmd)
@@ -722,7 +579,7 @@ example:
 
         if self._zk.exists(path):
             opts = map(lambda z: "%s/%s" % (path, z),
-                       self._zk.get_children(self._abspath(path)))
+                       self._zk.get_children(self.abspath(path)))
         elif "/" not in path:
             znodes = self._zk.get_children(self.curdir)
             opts = filter(lambda z: z.startswith(path), znodes)
